unit UMath;

interface

uses Math;

type

  //----------------------------------------------------------------------------
  TVector = record
    x:Double;
    y:Double;
    z:Double;
  end;

  //----------------------------------------------------------------------------
  TMatrix = record
    m:array[0..3,0..3]of Double;
  end;

  //----------------------------------------------------------------------------
  // Scalar
  function RoundOff(x:Double):integer;
  function Rad2Deg(r:Double):Double;
  function Deg2Rad(d:Double):Double;

  //----------------------------------------------------------------------------
  // Vector
  function VecInit(x,y,z:Double):TVector;
  function VecAdd(a,b:TVector):TVector;
  function VecSub(a,b:TVector):TVector;
  function VecLength(a:TVector):Double;
  function VecNormalize(a:TVector):TVector;
  function VecDotProduct(a,b:TVector):Double;
  function VecCrossProduct(a,b:TVector):TVector;

  //----------------------------------------------------------------------------
  // Matrix
  function MatInit(m00,m01,m02,m03,
                   m10,m11,m12,m13,
                   m20,m21,m22,m23,
                   m30,m31,m32,m33:Double):TMatrix;
  function MatIdentity():TMatrix;
  function MatMult(a,b:TMatrix):TMatrix;
  function MatApply(a:TMatrix;b:TVector):TVector;

  function MatRotX(r:Double):TMatrix;
  function MatRotY(r:Double):TMatrix;
  function MatRotZ(r:Double):TMatrix;
  function MatTranslate(x,y,z:Double):TMatrix;
  function MatScale(x,y,z:Double):TMatrix;

  //----------------------------------------------------------------------------
  // High level functions
  function MatYawPitchRoll(yaw,pitch,roll:Double):TMatrix;
  function MatProjection(fov,aspect,nearz,farz:Double):TMatrix;
  function MatLookAtView(from,at,up:TVector):TMatrix;

  //----------------------------------------------------------------------------

const
//  PI  = Pi;
  PI2 = Pi*2;

implementation

//----------------------------------------------------------------------------
function RoundOff(x:Double):integer;
begin
  if x>=0 then Result:=Trunc(x+0.5)
  else Result:=Trunc(x-0.5);
end;

function Rad2Deg(r:Double):Double;
begin
  Result:=(360/PI2*r);
{
  try
    Result:=(360/PI2*r);
  except
    Result:=0;
  end;
}
end;

function Deg2Rad(d:Double):Double;
begin
  Result:=(PI2/360*d);
{
  try
    Result:=(PI2/360*d);
  except
    Result:=0;
  end;
}
end;


//----------------------------------------------------------------------------
function VecInit(x,y,z:Double):TVector;
var v:TVector;
begin
  v.x := x;
  v.y := y;
  v.z := z;
  Result := v;
end;

function VecAdd(a,b:TVector):TVector;
begin
  Result := VecInit( (a.x + b.x), (a.y + b.y), (a.z + b.z) );
end;

function VecSub(a,b:TVector):TVector;
begin
  Result := VecInit( (a.x - b.x), (a.y - b.y), (a.z - b.z) );
end;

function VecLength(a:TVector):Double;
begin
  Result := Sqrt( a.x*a.x + a.y*a.y + a.z*a.z );
end;

function VecNormalize(a:TVector):TVector;
var norm:Double;
begin
//  norm := 1/VecLength(a);

  norm := VecLength(a);
  if 0<>norm then begin
    norm := 1/norm;
  end else begin
    norm := 1;
  end;

  Result := VecInit( (a.x * norm), (a.y * norm), (a.z * norm) );
end;

function VecDotProduct(a,b:TVector):Double;
begin
  Result := ( a.x*b.x + a.y*b.y + a.z*b.z );
end;

function VecCrossProduct(a,b:TVector):TVector;
begin
  Result := VecInit(
    (a.y*b.z - a.z*b.y),
    (a.z*b.x - a.x*b.z),
    (a.x*b.y - a.y*b.x)
  );
end;


//----------------------------------------------------------------------------
function MatInit(m00,m01,m02,m03,
                 m10,m11,m12,m13,
                 m20,m21,m22,m23,
                 m30,m31,m32,m33:Double):TMatrix;
var m:TMatrix;
begin
  m.m[0,0]:=m00; m.m[0,1]:=m01; m.m[0,2]:=m02; m.m[0,3]:=m03;
  m.m[1,0]:=m10; m.m[1,1]:=m11; m.m[1,2]:=m12; m.m[1,3]:=m13;
  m.m[2,0]:=m20; m.m[2,1]:=m21; m.m[2,2]:=m22; m.m[2,3]:=m23;
  m.m[3,0]:=m30; m.m[3,1]:=m31; m.m[3,2]:=m32; m.m[3,3]:=m33;
  Result := m;
end;

function MatIdentity():TMatrix;
begin
  Result := MatInit(
    1,0,0,0,
    0,1,0,0,
    0,0,1,0,
    0,0,0,1
  );
end;

function MatMult(a,b:TMatrix):TMatrix;
begin
  Result := MatInit(
    a.m[0,0]*b.m[0,0] + a.m[0,1]*b.m[1,0] + a.m[0,2]*b.m[2,0] + a.m[0,3]*b.m[3,0],
    a.m[0,0]*b.m[0,1] + a.m[0,1]*b.m[1,1] + a.m[0,2]*b.m[2,1] + a.m[0,3]*b.m[3,1],
    a.m[0,0]*b.m[0,2] + a.m[0,1]*b.m[1,2] + a.m[0,2]*b.m[2,2] + a.m[0,3]*b.m[3,2],
    a.m[0,0]*b.m[0,3] + a.m[0,1]*b.m[1,3] + a.m[0,2]*b.m[2,3] + a.m[0,3]*b.m[3,3],
    a.m[1,0]*b.m[0,0] + a.m[1,1]*b.m[1,0] + a.m[1,2]*b.m[2,0] + a.m[1,3]*b.m[3,0],
    a.m[1,0]*b.m[0,1] + a.m[1,1]*b.m[1,1] + a.m[1,2]*b.m[2,1] + a.m[1,3]*b.m[3,1],
    a.m[1,0]*b.m[0,2] + a.m[1,1]*b.m[1,2] + a.m[1,2]*b.m[2,2] + a.m[1,3]*b.m[3,2],
    a.m[1,0]*b.m[0,3] + a.m[1,1]*b.m[1,3] + a.m[1,2]*b.m[2,3] + a.m[1,3]*b.m[3,3],
    a.m[2,0]*b.m[0,0] + a.m[2,1]*b.m[1,0] + a.m[2,2]*b.m[2,0] + a.m[2,3]*b.m[3,0],
    a.m[2,0]*b.m[0,1] + a.m[2,1]*b.m[1,1] + a.m[2,2]*b.m[2,1] + a.m[2,3]*b.m[3,1],
    a.m[2,0]*b.m[0,2] + a.m[2,1]*b.m[1,2] + a.m[2,2]*b.m[2,2] + a.m[2,3]*b.m[3,2],
    a.m[2,0]*b.m[0,3] + a.m[2,1]*b.m[1,3] + a.m[2,2]*b.m[2,3] + a.m[2,3]*b.m[3,3],
    a.m[3,0]*b.m[0,0] + a.m[3,1]*b.m[1,0] + a.m[3,2]*b.m[2,0] + a.m[3,3]*b.m[3,0],
    a.m[3,0]*b.m[0,1] + a.m[3,1]*b.m[1,1] + a.m[3,2]*b.m[2,1] + a.m[3,3]*b.m[3,1],
    a.m[3,0]*b.m[0,2] + a.m[3,1]*b.m[1,2] + a.m[3,2]*b.m[2,2] + a.m[3,3]*b.m[3,2],
    a.m[3,0]*b.m[0,3] + a.m[3,1]*b.m[1,3] + a.m[3,2]*b.m[2,3] + a.m[3,3]*b.m[3,3]
  );
end;

function MatApply(a:TMatrix; b:TVector):TVector;
begin
  Result := VecInit(
    a.m[0,0]*b.x + a.m[1,0]*b.y + a.m[2,0]*b.z + a.m[3,0],
    a.m[0,1]*b.x + a.m[1,1]*b.y + a.m[2,1]*b.z + a.m[3,1],
    a.m[0,2]*b.x + a.m[1,2]*b.y + a.m[2,2]*b.z + a.m[3,2]
  );
end;

function MatRotX(r:Double):TMatrix;
var s,c:Double;
begin
  s := Sin(r);
  c := Cos(r);
  Result := MatInit(
    1, 0, 0, 0,
    0, c, s, 0,
    0,-s, c, 0,
    0, 0, 0, 1
  );
end;

function MatRotY(r:Double):TMatrix;
var s,c:Double;
begin
  s := Sin(r);
  c := Cos(r);
  Result := MatInit(
    c, 0,-s, 0,
    0, 1, 0, 0,
    s, 0, c, 0,
    0, 0, 0, 1
  );
end;

function MatRotZ(r:Double):TMatrix;
var s,c:Double;
begin
  s := Sin(r);
  c := Cos(r);
  Result := MatInit(
    c, s, 0, 0,
   -s, c, 0, 0,
    0, 0, 1, 0,
    0, 0, 0, 1
  );
end;

function MatTranslate(x,y,z:Double):TMatrix;
begin
  Result := MatInit(
    1, 0, 0, 0,
    0, 1, 0, 0,
    0, 0, 1, 0,
    x, y, z, 1
  );
end;

function MatScale(x,y,z:Double):TMatrix;
begin
  Result := MatInit(
    x, 0, 0, 0,
    0, y, 0, 0,
    0, 0, z, 0,
    0, 0, 0, 1
  );
end;


//----------------------------------------------------------------------------
function MatYawPitchRoll(yaw,pitch,roll:Double):TMatrix;
begin
  Result := MatMult( MatMult( MatRotY(yaw), MatRotX(pitch) ), MatRotZ(roll) );
end;

function MatLookAtView(from,at,up:TVector):TMatrix;
var F,S,U:TVector;
begin
  F := VecNormalize(VecSub(at,from));       //front
  S := VecNormalize(VecCrossProduct(up,F)); //side
  U := VecCrossProduct(F,S);                //up
  Result := MatInit(
    S.x, U.x, F.x, 0,
    S.y, U.y, F.y, 0,
    S.z, U.z, F.z, 0,
    -VecDotProduct( S, from ),
    -VecDotProduct( U, from ),
    -VecDotProduct( F, from ),
    1
  );
end;

function MatProjection( fov,aspect,nearz,farz:Double ):TMatrix;
var w,h,f,n:Double;
begin
  h := (1/Tan(fov*0.5));
  w := (h/aspect);
  f := farz/(farz-nearz);
  n := -nearz*f;
  Result := MatInit(
    w, 0, 0, 0,
    0, h, 0, 0,
    0, 0, f, 1,
    0, 0, n, 0
  );
{
  try
    h := (1/Tan(fov*0.5));
    w := (h/aspect);
    f := farz/(farz-nearz);
    n := -nearz*f;
    Result := MatInit(
      w, 0, 0, 0,
      0, h, 0, 0,
      0, 0, f, 1,
      0, 0, n, 0
    );
  except
    Result := MatIdentity;
  end;
}
end;









end.
