unit URender;

interface

uses
  Windows, Graphics,
  UMath;

type

  //----------------------------------------------------------------------------
  // Rendering Manager
  TRender = class
  protected
    OffScreen:TBitmap;
    CenterX:Double;
    CenterY:Double;

    Master:TMatrix;
    World:TMatrix;
    View:TMatrix;
    Proj:TMatrix;

  public
    constructor Create( w, h:integer );
    destructor Destroy; override;
    procedure Clear;
    procedure Present( screen:HDC; rect:TRECT );

    procedure SetMaster( m:TMatrix );
    procedure SetWorld( m:TMatrix );
    procedure SetView( m:TMatrix );
    procedure SetProj( m:TMatrix );

    procedure DrawPoly( v0, v1, v2:TVector );
    procedure DrawPrim( v:array of TVector );
    procedure DrawIdxPrim( v:array of TVector; idx:array of integer );

  end;

  //----------------------------------------------------------------------------
  // Scene base
  TScene = class
    constructor Create( var rndr:TRender );
  protected
    Renderer:^TRender;
  public
    function Init:boolean;virtual;abstract;
    procedure Update( delta:Double );virtual;abstract;
  end;

  //----------------------------------------------------------------------------

const
  //-------------------------------------------------
  // test triangle
  tri_vtx:array[0..2]of TVector = (
    ( x:  0; y: 50; z: 0 ),
    ( x: 25; y:  0; z: 0 ),
    ( x:-25; y:  0; z: 0 )
  );
  tri_idx:array[0..2]of integer = (
    0,1,2
  );

  //-------------------------------------------------
  // test box
  box_vtx:array[0..7]of TVector = (
    ( x: 25; y: 25; z: 25 ),
    ( x:-25; y: 25; z: 25 ),
    ( x: 25; y:-25; z: 25 ),
    ( x:-25; y:-25; z: 25 ),

    ( x: 25; y: 25; z:-25 ),
    ( x:-25; y: 25; z:-25 ),
    ( x: 25; y:-25; z:-25 ),
    ( x:-25; y:-25; z:-25 )
  );
  box_idx:array[0..35]of integer = (
    0,1,2, 2,3,1,
    4,5,7, 7,6,4,
    4,5,0, 0,1,5,
    2,3,6, 6,7,3,
    0,2,6, 6,4,0,
    1,3,5, 5,7,3
  );

  //-------------------------------------------------

implementation

//----------------------------------------------------------------------------
constructor TRender.Create( w,h:integer );
begin
  OffScreen := TBitmap.Create;
  OffScreen.Width  := w;
  OffScreen.Height := h;
  CenterX := w / 2; 
  CenterY := h / 2;

  Master := MatIdentity;
  World  := MatIdentity;
  View   := MatIdentity;
  Proj   := MatIdentity;

end;

destructor TRender.Destroy;
begin
  OffScreen.Free;
end;

procedure TRender.Clear;
begin
  BitBlt( OffScreen.Canvas.Handle,0,0,OffScreen.Width,OffScreen.Height, 0,0,0, WHITENESS );
end;

procedure TRender.Present(screen:HDC;rect:TRECT);
begin
  StretchBlt( screen,rect.Left,rect.Top,rect.Right,rect.Bottom,
              OffScreen.Canvas.Handle,0,0,OffScreen.Width,OffScreen.Height,
              SRCCOPY );
end;

procedure TRender.SetMaster(m:TMatrix);
begin
  Master := m;
end;

procedure TRender.SetWorld( m:TMatrix );
begin
  World  := m;
  Master := MatMult(MatMult(World,View),Proj);
end;

procedure TRender.SetView( m:TMatrix );
begin
  View   := m;
  Master := MatMult(MatMult(World,View),Proj);
end;

procedure TRender.SetProj( m:TMatrix );
begin
  Proj   := m;
  Master := MatMult(MatMult(World,View),Proj);
end;

procedure TRender.DrawPoly( v0,v1,v2:TVector );

  function Transform(v:TVector):TPoint;
  var z:Double;
  begin
    z:=1;
    if v.z>0 then z:=(1/v.z);
    Result.x := RoundOff( CenterX + v.x * z * CenterX );
    Result.y := RoundOff( CenterY - v.y * z * CenterY );
  end;

var
//  pt:array[0..2]of TPoint;
  pt:array[0..3]of TPoint;
begin
  pt[0] := Transform(v0);
  pt[1] := Transform(v1);
  pt[2] := Transform(v2);
  pt[3] := pt[0];
//  OffScreen.Canvas.Polygon( pt );
  OffScreen.Canvas.Polyline( pt );
end;

procedure TRender.DrawPrim( v:array of TVector );
var
  i:integer;
  tv0,tv1,tv2:TVector;
begin
  i:=0;
  while i<=High(v) do
  begin
    tv0 := MatApply(Master,v[i+0]);
    tv1 := MatApply(Master,v[i+1]);
    tv2 := MatApply(Master,v[i+2]);
    DrawPoly(tv0,tv1,tv2);
    inc(i,3);
  end;
end;

procedure TRender.DrawIdxPrim( v:array of TVector; idx:array of integer );
var
  i:integer;
  tv0,tv1,tv2:TVector;
begin
  i:=0;
  while i<=High(idx) do
  begin
    tv0 := MatApply(Master,v[idx[i+0]]);
    tv1 := MatApply(Master,v[idx[i+1]]);
    tv2 := MatApply(Master,v[idx[i+2]]);
    DrawPoly(tv0,tv1,tv2);
    inc(i,3);
  end;
end;



//----------------------------------------------------------------------------
constructor TScene.Create( var rndr:TRender );
begin
  Renderer := @rndr;
  Init;
end;




end.
